@@ -24,10 +24,20 @@ import typing
 from typing import Union, List, Iterator
 from threading import Thread
 import queue
+from .logger import Logger
 
+logger = Logger("pyal.algorithm")
 INF = math.inf
 EPSILON = 1e-6
 
+def read_file_content(file_name: str, mode="r"):
+  with open(file_name, mode) as fin:
+    return fin.read().strip()
+
+def read_file_lines(file_name: str, mode="r"):
+  with open(file_name, mode) as fin:
+    for ln in fin:
+      yield ln.rstrip()
 
 def load_py_data(py_file):
   user_data = {}
@@ -36,7 +46,7 @@ def load_py_data(py_file):
       exec(compile(fin.read(), "py_data", "exec"), user_data)
       return user_data
     except Exception as error:
-      Logger.error(error)
+      logger.error(error)
       return {}
 
 
@@ -157,11 +167,11 @@ def csv_file_read(file_name, max_num: int=-1)-> typing.Iterator:
         break
 
       if data_num > 0 and data_num % 10_000 == 0:
-        Logger.info(f"{file_name}: {data_num} lines have been loaded.")
+        logger.info(f"{file_name}: {data_num} lines have been loaded.")
 
       yield row
 
-  Logger.info(f"{file_name}: #data={data_num:,}")
+  logger.info(f"{file_name}: #data={data_num:,}")
 
 def csv_file_write(data: typing.Iterator, field_names: list,
                    file_name, remove_extra_keys=True, **kwargs):
@@ -188,7 +198,7 @@ def pydict_file_read(file_name, max_num: int = -1) -> typing.Iterator:
       if max_num >= 0 and idx + 1 > max_num:
         break
       if idx > 0 and idx % 10_000 == 0:
-        Logger.info(f"{file_name}: {idx} lines have been loaded.")
+        logger.info(f"{file_name}: {idx} lines have been loaded.")
 
       try:
         obj = eval(ln)
@@ -196,9 +206,9 @@ def pydict_file_read(file_name, max_num: int = -1) -> typing.Iterator:
         data_num += 1
 
       except Exception as err:
-        Logger.error(f"reading {file_name}:{idx + 1}: {err} '{ln}'")
+        logger.error(f"reading {file_name}:{idx + 1}: {err} '{ln}'")
 
-  Logger.info(f"{file_name}: #data={data_num:,}")
+  logger.info(f"{file_name}: #data={data_num:,}")
 
 
 def json_file_write(data, file_name: str):
@@ -228,16 +238,16 @@ def pydict_file_write(data: typing.Iterator, file_name: str, **kwargs) -> None:
       num += 1
       obj_str = str(obj)
       if "\n" in obj_str:
-        Logger.error(f"pydict_file_write: not '\\n' is allowed: '{obj_str}'")
+        logger.error(f"pydict_file_write: not '\\n' is allowed: '{obj_str}'")
       print(obj, file=fou)
       if kwargs.get("print_log", True) and num % 10_000 == 0:
-        Logger.info(f"{file_name} has been written {num} lines")
+        logger.info(f"{file_name} has been written {num} lines")
       if kwargs.get("flush_freq", None) is not None and \
         num % kwargs["flush_freq"] == 0:
         fou.flush()
 
   if kwargs.get("print_log", True):
-    Logger.info(f"{file_name} has been written {num} lines totally")
+    logger.info(f"{file_name} has been written {num} lines totally")
 
 
 def get_file_extension(file_name: str) -> str:
@@ -303,51 +313,6 @@ def to_readable_time(seconds: float):
   return " ".join(result)
 
 
-def __strdate(timezone: str, now):
-  city = timezone.split("/")[-1]
-  ts = now.strftime("%Y-%m-%d_%Ih-%Mm-%Ss_%p")
-  return f"{city}_{ts}"
-
-
-def get_log_time(utc_time: bool = True, country_city: str = None):
-  '''
-  utc_time: if False, return local time(server);
-            if True, return local time(city).
-  country_city : When utc_time is true,  if city is None, return UTC(0).
-                See pytz/__init__.py:510, all_timezones
-
-  e.g., SF time is UTC+8, then get_log_time(True) - 8 = get_log_time(False)
-  '''
-  if utc_time:
-    if is_none_or_empty(country_city):
-      now = datetime.datetime.utcnow()
-      return __strdate("utc", now)
-    else:
-      now = datetime.datetime.now(pytz.timezone(country_city))
-      return __strdate(country_city, now)
-
-  else:
-    now = datetime.datetime.now()
-    return __strdate("local", now)
-
-
-def get_future_time(days=0,
-                    hours=0,
-                    minutes=0,
-                    seconds=0,
-                    country_city: str = None):
-  delta = datetime.timedelta(days=days,
-                             hours=hours,
-                             minutes=minutes,
-                             seconds=seconds)
-  if is_none_or_empty(country_city):
-    finished_time = datetime.datetime.now() + delta
-    return __strdate("utc", finished_time)
-  else:
-    finished_time = datetime.datetime.now(pytz.timezone(country_city)) + delta
-    return __strdate(country_city, finished_time)
-
-
 @functools.lru_cache
 def get_IPs():
   return set(
@@ -394,10 +359,10 @@ def command(cmd: str,
       account = os.getlogin()
     full_cmd = f"ssh -oStrictHostKeyChecking=no {account}@{server_ip} '{cmd}'"
 
-  Logger.debug(f"[start] executing '{full_cmd}'")
+  logger.debug(f"[start] executing '{full_cmd}'")
   result = subprocess.run(full_cmd, shell=True, capture_output=capture_output)
   status = "OK" if result.returncode == 0 else "fail"
-  Logger.debug(f"[finish - {status}] '{full_cmd}'")
+  logger.debug(f"[finish - {status}] '{full_cmd}'")
 
   if capture_output:
     return result.returncode, result.stdout.decode(), result.stderr.decode()
@@ -410,13 +375,13 @@ def to_utf8(line) -> typing.Union[str, None]:
     try:
       return line.encode("utf8")
     except:
-      Logger.warn("in toUtf8(...)")
+      logger.warn("in toUtf8(...)")
       return None
 
   elif type(line) is bytes:
     return line
 
-  Logger.error("wrong type in toUtf8(...)")
+  logger.error("wrong type in toUtf8(...)")
   return None
 
 
@@ -430,7 +395,7 @@ def print_flush(cont, stream=None) -> None:
 def display_server_info():
   host_name = socket.gethostname()
   ip = get_server_ip()
-  Logger.info(f"server information: {host_name}({ip}), process: {os.getpid()}")
+  logger.info(f"server information: {host_name}({ip}), process: {os.getpid()}")
 
 
 def get_available_gpus(server_ip=None, account=None):
@@ -441,10 +406,10 @@ def get_available_gpus(server_ip=None, account=None):
                   capture_output=True,
                   server_ip=server_ip,
                   account=account)[1]
-    Logger.debug(f"server: {server_ip}, {res}")
+    logger.debug(f"server: {server_ip}, {res}")
     res = res.split("\n")
     if len(res) <= 6:
-      Logger.error(
+      logger.error(
           f"can not obtain correct nvidia-smi result: {' '.join(res)}")
       yield -1
       return
@@ -466,10 +431,10 @@ def get_available_gpus(server_ip=None, account=None):
     ret = timeout(find_all, [], 30)
     return ret
   except TimeoutError:
-    Logger.error(f"Time out: get_available_gpus({server_ip})")
+    logger.error(f"Time out: get_available_gpus({server_ip})")
     return []
   except Exception as error:
-    Logger.error(error)
+    logger.error(error)
     return []
 
 
@@ -516,76 +481,18 @@ class Timer(object):
 
   def __enter__(self) -> None:
     if not is_none_or_empty(self.title):
-      Logger.info(f"Timer starts:\t '{self.title}'")
+      logger.info(f"Timer starts:\t '{self.title}'")
     self.__starting = time.time()
     return self
 
   def __exit__(self, *args) -> None:
     self.__duration = time.time() - self.__starting
     if not is_none_or_empty(self.title):
-      Logger.info(
+      logger.info(
           f"Timer finishes:\t '{self.title}', takes {to_readable_time(self.duration)} "
           f"seconds.")
 
 
-class Logger:
-  '''
-  debug=0, info=1, warning=2, error=3
-  '''
-  level = 1
-  outstream = sys.stdout
-  country_city = ""  #"Asia/Chongqing", 'America/Los_Angeles'
-
-  @staticmethod
-  def reset_outstream(out_file: str, append=False):
-    mode = "a" if append else "w"
-    Logger.outstream = open(out_file, mode)
-
-  @staticmethod
-  def set_level(level):
-    Logger.level = level
-
-  @staticmethod
-  def is_debug():
-    return Logger.level <= 0
-
-  @staticmethod
-  def debug(*args):
-    if Logger.level <= 0:
-      print(get_log_time(country_city=Logger.country_city),
-            "DEBUG:",
-            *args,
-            file=Logger.outstream)
-      Logger.outstream.flush()
-
-  @staticmethod
-  def info(*args):
-    if Logger.level <= 1:
-      print(get_log_time(country_city=Logger.country_city),
-            "INFO:",
-            *args,
-            file=Logger.outstream)
-      Logger.outstream.flush()
-
-  @staticmethod
-  def warn(*args):
-    if Logger.level <= 2:
-      print(get_log_time(country_city=Logger.country_city),
-            "WARN:",
-            *args,
-            file=Logger.outstream)
-      Logger.outstream.flush()
-
-  @staticmethod
-  def error(*args):
-    if Logger.level <= 3:
-      print(get_log_time(country_city=Logger.country_city),
-            "ERR:",
-            *args,
-            file=Logger.outstream)
-      Logger.outstream.flush()
-
-
 def bit_not(m):
   import ctypes
   return ctypes.c_uint32(~m).value
@@ -1023,7 +930,26 @@ class Pool:
   def __init__(self, processes=4):
     self._worker_num = processes
 
+  def map1(self, user_func, args_list: Union[List, Iterator])-> Iterator:
+    '''
+    :param user_func:
+    :param args_list:
+    :return: fake stream-based results
+    '''
+    from concurrent.futures import ThreadPoolExecutor
+
+    with ThreadPoolExecutor(max_workers=self._worker_num) as executor:
+      results = executor.map(lambda args: user_func(*args), args_list)
+
+    for result in results:
+      yield result
+
   def map(self, user_func, args_list: Union[List, Iterator])-> Iterator:
+    '''
+    :param user_func:
+    :param args_list:
+    :return: stream-based results
+    '''
     def _thread_worker(index_queue: queue.Queue, out_queue: queue.Queue):
       while True:
         data_id = index_queue.get()
@@ -1054,3 +980,6 @@ class Pool:
         yield results[cur_index]
         del results[cur_index]
         cur_index += 1
+
+    for td in threads:
+      td.join()
